from peyetribe import EyeTribe
import time
import scipy.io
import numpy as np
import os
import pygame
from pygame.locals import *

WIDTH           = 1280
HEIGHT          = 800
ORIGINAL_IMAGE  = 'money1.jpg'
IMAGES_DIR      = './stack/'
DONE            = False
TOGGLE          = False
NUM_READINGS    = 4
LAST_ORIGINAL   = False
count = 0

'''
Description - Loads the centres and stack_mask.

args - takes in no arguments

returns - centres, stack_mask
    centres: Matrix with coordinates of the centres for the different blurs
    stack_mask: Matrix of the size of the image that maps each pixel to the
    index of the image in the stack to be displayed if the gaze is fixed at
    the said pixel 
'''
def preprocessed():
    centres = scipy.io.loadmat('centres.mat')['centres']
    print('No. of centres: ')
    print(np.shape(centres)[1])
    stack_mask = scipy.io.loadmat('stack_mask.mat')['stack_mask']
    print('Size of stack mask: ')
    print(np.shape(stack_mask))
    return centres, stack_mask

'''
Description - Loads the preprocessed (blurred iages).

args - path, ncentres
    path: path to the saved preprocessed/blurreed images
    ncentres: the number of blurred images

returns - image_list
    image_list: a list of all the blurred images loaded into memory
'''
def loadImages(path, ncentres):
    image_list = []
    for i in xrange(ncentres):
        filename = str(i+1) + '.png'
        filepath = os.path.join(path, filename)
        im = pygame.image.load(filepath)
        image_list.append(im)
    return image_list

'''
Description - Reads the X and Y coordinates of the viewer's gaze from the tracker. 
Returns co-ordinates after removing noise and smoothing. Makes certain that the 
value being returned cooresponds to a FIXATION and not a SACCADE. By editing the
global variable NUMREADINGS, the extent of smoothing can be varied.

Note - Increasing the NUMREADINGS may lead to poorer performance due to higher
latency.

args - tracker
    tracker: the tracker instance of the running eyetribe tracker

returns - X, Y
    X, Y: the X and Y coordinates of the viewer's gaze
'''
def getXY(tracker):
    X = 0.0
    Y = 0.0
    readings = 0
    while readings<NUM_READINGS:
        n = tracker.next()
        fix = False
        n_list = str(n).split(';')
        if n_list[3] == 'F':
            fix = True
        if fix:
            y_pos, x_pos = n_list[7:9]
            x_pos, y_pos = int(round(float(x_pos))), int(round(float(y_pos)))
            x_pos = max(0, min(HEIGHT-1, x_pos))
            y_pos = max(0, min(WIDTH-1, y_pos))
            X += x_pos
            Y += y_pos
            readings += 1
    X = int(round(X/NUM_READINGS))
    Y = int(round(Y/NUM_READINGS))
    return X, Y

'''
Description -
1. Creates a tracker object
2. Creates a rendering environment
3. Loades pre-processed/blurred images, centres and stack_mask
4. Initialises the rendering untill 'ESC' key is pressed
5. Gets X and Y co-ordinates of viewer's gaze
6. Displays based on TOGGLE state:
    (i) State 1: Display original unblurred image with a red circle at the 
    location of viewer's gaze
    (ii) State 2: Display blurred image corresponding to the location of
    viewer's gaze
The TOGGLE state can be switched by pressing any key except the ESC key
7. Press ESC to end the experiment; close the rendering environment and
the link to the tracker
'''
if __name__ == "__main__":

#    Instantiate tracker
    tracker = EyeTribe()
    tracker.connect()
    n = tracker.next()
    tracker.pushmode()

#    Initialise game environment
    pygame.init()
#    windowSurface = pygame.display.set_mode((WIDTH, HEIGHT), pygame.FULLSCREEN)
    windowSurface = pygame.display.set_mode((WIDTH, HEIGHT), 0, 32)
    pygame.display.set_caption('Foveated rendering')

#    Load images and centres and stack mask
    orig_img = pygame.image.load(ORIGINAL_IMAGE)
    centres, stack_mask = preprocessed()
    images_list = loadImages(IMAGES_DIR, np.shape(centres)[1])

#    Rendering
    while not DONE:
        count += 1
        events = pygame.event.get()
        for event in events:
            if event.type == pygame.KEYDOWN:
                if event.key == pygame.K_ESCAPE:
                    DONE = True
                    break
                else:
                    TOGGLE = not TOGGLE
            elif event.type == pygame.QUIT:
                DONE = True
                break
        if DONE:
            pygame.quit()
            break

#            Get coordinates
        x_pos, y_pos = getXY(tracker)

        if TOGGLE:
#            x_pos, y_pos = 100, 100
            image_index = stack_mask[x_pos][y_pos]
            if count%100 == 0:
                print 'Displaying for count: ', count
                print 'X: ', x_pos, ' Y: ', y_pos
                print '\n'
                print 10*'='
                print '\n'
            windowSurface.blit(images_list[image_index], (0, 0))
            pygame.display.flip()
            pygame.draw.circle(windowSurface, (255, 0, 0), (y_pos, x_pos), 20, 3)
            pygame.display.update()
            LAST_ORIGINAL = False
#            pygame.time.wait(33)
        else:
            if not LAST_ORIGINAL:
                windowSurface.blit(orig_img, (0, 0))
                pygame.display.flip()
            pygame.draw.circle(windowSurface, (255, 0, 0), (y_pos, x_pos), 25, 2)
            pygame.display.update()
            LAST_ORIGINAL = True

    pygame.quit()

    tracker.pullmode()
    tracker.close()